#include "config.h"
#include "structs.h"
#include <EEPROM.h>
#include "EEPROM_settings.h"

void WriteSection(uInt address,float data);
float ReadSection(uInt address);

void setup() {
  pinMode(driver_X::DIRECTION, OUTPUT);
  pinMode(driver_Z::DIRECTION, OUTPUT);
  pinMode(driver_X::STEP, OUTPUT);
  pinMode(driver_Z::STEP, OUTPUT);
  pinMode(spindle::INDEX, INPUT);
  pinMode(spindle::SYNC, INPUT);
  Serial.begin(115200,SERIAL_8N1);
  Serial.println("\nGrbl 1.1g ['$' for help]");
  Serial.println("[MSG:CNC by skmskm ['$' cry for help]]\n");
  if( EEPROM.read(EEPROM.end()-1) != 0 )
  {
    EEPROM.write(EEPROM.end()-1,0);
    for (int i = 0 ; i < EEPROM.length() ; i++) {
      EEPROM.write(i, 0);
    }
    //write config
    WriteSection(100*sizeof(float),driver_X::STEP_PERMM);
    WriteSection(100*sizeof(float),driver_Y::STEP_PERMM);
    WriteSection(100*sizeof(float),driver_Z::STEP_PERMM);
    WriteSection(100*sizeof(float),driver_A::STEP_PERANG);
    WriteSection(100*sizeof(float),driver_B::STEP_PERANG);
    WriteSection(100*sizeof(float),driver_T::STEP_PERPOS);
    for( int i = 110; i < 116; i++)
      WriteSection(i*sizeof(float),2000);
    for( int i = 120; i < 126; i++)
      WriteSection(i*sizeof(float),500);
    for( int i = 130; i < 136; i++)
      WriteSection(i*sizeof(float),500);
  }
}


float GetFloat(int &pos);
Procesor::sFrame* ProcessFrame(int pos = 0);
float* FindInFrame(char letter);
void(* resetFunc) (void) = 0;//declare reset function at address 0
void DoMove();

void loop()
{
  while ((Serial.available() > 0) && !Procesor::data_ready ){
    int ch = Serial.read();
    if(( ch == 10 ) || ( ch == 13 ))
    {
      Procesor::data_ready = true;
    }
    else
    {
      switch(ch)
      {
        case CMD_CYCLE_START:
          break;
        case CMD_FEED_HOLD:
          if( Machine::machine_status == 2 )
          {
            Machine::machine_status = 3;
            Machine::pause::copy_TCCR0B = TCCR0B;
            Machine::pause::copy_TCCR1B = TCCR1B;
            Machine::pause::copy_TCCR2B = TCCR2B;
            Machine::pause::copy_TCCR3B = TCCR3B;
            Machine::pause::copy_TCCR4B = TCCR4B;
            TCCR0B = 0;
            TCCR1B = 0;
            TCCR2B = 0;
            TCCR3B = 0;
            TCCR4B = 0;
          }
          if( Machine::machine_status == 3 )
          {
            TCCR0B = Machine::pause::copy_TCCR0B;
            TCCR1B = Machine::pause::copy_TCCR1B;
            TCCR2B = Machine::pause::copy_TCCR2B;
            TCCR3B = Machine::pause::copy_TCCR3B;
            TCCR4B = Machine::pause::copy_TCCR4B;
            Machine::machine_status = 1;
          }
          break;
        case CMD_RESET:
          resetFunc();
          break;
        case CMD_STATUS_REPORT:
          Serial.print("<");
          if( Machine::machine_status == 0 )
            Serial.print("Idle");
          if( Machine::machine_status == 1 )
            Serial.print("Jog");
          if( Machine::machine_status == 2 )
            Serial.print("Run");
          if( Machine::machine_status == 3 )
            Serial.print("Hold");
          Serial.print("|MPos:");
          Serial.print(Machine::pos_X);
          Serial.print(",");
          Serial.print(Machine::pos_Y);
          Serial.print(",");
          Serial.print(Machine::pos_Z);
          Serial.println("|FS:0,0|Ov:100,100,100>");
          break;
        case CMD_JOG_CANCEL:
          Machine::machine_status = 0;
          Machine::pos_X -= Machine::dirX*(float)TIMERS::counterX_done/(float)driver_X::STEP_PERMM;
          Machine::pos_Y -= Machine::dirY*(float)TIMERS::counterY_done/(float)driver_Y::STEP_PERMM;
          Machine::pos_Z -= Machine::dirZ*(float)TIMERS::counterZ_done/(float)driver_Z::STEP_PERMM;
          TIMERS::counterX_target = 0;
          TIMERS::counterZ_target = 0;
          TCCR2B = 0;
          TCCR1B = 0;
          break;
        default:
          if( ch < 0x86 )
            Procesor::buff[Procesor::buff_pos++]=ch;
      }
    }
  }
  if(( Procesor::data_ready ) && ( Procesor::buff_pos == 0 ))
    Procesor::data_ready = false;
  Procesor::buff[Procesor::buff_pos] = 0;
  if( Machine::machine_status == 1 )//joging loop
  {
    if((TIMERS::counterX_target == 0) && (TIMERS::counterY_target == 0) && (TIMERS::counterZ_target == 0))
    {
        Machine::pos_X -= Machine::dirX*(float)TIMERS::counterX_done/(float)driver_X::STEP_PERMM;
        Machine::pos_Y -= Machine::dirY*(float)TIMERS::counterY_done/(float)driver_Y::STEP_PERMM;
        Machine::pos_Z -= Machine::dirZ*(float)TIMERS::counterZ_done/(float)driver_Z::STEP_PERMM;
        Machine::machine_status = 0;
    }
  }
  if( Machine::delay_time )
  {
      unsigned long t = millis() - Machine::timems;
      Machine::time -= t;
      if( Machine::time <= 0 )
      {
        Machine::delay_time = false;
        Machine::machine_status = 0;
      }
  }
  if( Machine::machine_status == 2 )//run
  {
    if( Machine::sub_mode == 33 )//G33
    {
      //syncing for G33
      if( TIMERS::counterZ_target == 0 )
      {
        Machine::pos_Z = Procesor::Zvalue;
        Machine::machine_status = 0;//idle
      }
    }
    if( Machine::sub_mode == 1 )
    {
      if( (TIMERS::counterX_target == 0) && (TIMERS::counterY_target == 0) && (TIMERS::counterZ_target == 0) )
      {
        Machine::pos_X = Procesor::Xvalue;
        Machine::pos_Y = Procesor::Yvalue;
        Machine::pos_Z = Procesor::Zvalue;
        Machine::machine_status = 0;//idle
      }
    }
  }
  if( Procesor::data_ready )//check for $ commands
  {
    Procesor::len = strlen(Procesor::buff);
    if( Procesor::buff[0] == 'S' )
    {
      int pos = 1;
      float value = GetFloat(pos);
      Machine::speed_S = value;
      Procesor::buff_pos = 0;
      Procesor::data_ready = false;
    }
    if(( Procesor::buff[0] == 'M' ) && isDigit(Procesor::buff[1]))
    {
      int pos = 1;
      float value = GetFloat(pos);
      if( value == 5 )//M5 stop spindle
      {
        Machine::M3_M5 = 0;//off
        TCCR5B = 0;
        digitalWrite(spindle::ENABLE, LOW);
      }
      if(( value == 3 ) || ( value == 4 ))//M3/4 stop spindle
      {
        Procesor::sFrame * frame = ProcessFrame();
        float * f;
        f = FindInFrame('S');
        Machine::speed_S = (f != 0 )?f[0]:Machine::speed_S;
        if( Machine::speed_S > 0 )
        {
          digitalWrite(spindle::ENABLE, HIGH);
          if( value == 3 )
          {
            digitalWrite(spindle::DIRECTION, spindle::FLIP);
            Machine::M3_M5 = 1;//on
          }
          if( value == 4 )
          {
            digitalWrite(spindle::DIRECTION, !spindle::FLIP);
            Machine::M3_M5 = 2;//on
          }
          TCCR5B = (1<<WGM12) | (1<<CS12); 
          TIMSK5 = (1<<OCIE1A);
          float Actual_speed = (Machine::speed_S - spindle::MIN_SPEED);
          Actual_speed /= (float)(spindle::MAX_SPEED - spindle::MIN_SPEED);
          Actual_speed *= (float)(spindle::MAX_SPEED_PWM_VALUE - spindle::MIN_SPEED_PWM_VALUE);
          TIMERS::spindle_speed = (uInt)(Actual_speed*100.0F);
          TIMERS::spindle_up_down = false;
          OCR5A = TIMERS::spindle_speed;
        }
      }
      Procesor::buff_pos = 0;
      Procesor::data_ready = false;
    }
    
    if( Procesor::buff[0] == '$' )
    {
      Procesor::buff_pos = 0;
      Procesor::data_ready = false;
      if( Procesor::buff[1] == '$')
      {
        int field = 0;
        while(settings::config[field].ID != 0)
        {
          Serial.print("$");
          Serial.print(settings::config[field].ID);
          Serial.print("=");
          Serial.print(ReadSection(settings::config[field].ID*sizeof(float)));
          Serial.print(" ");
          Serial.print(settings::config[field].str);
          field++;
        }
        
      }
      if( isDigit(Procesor::buff[1]) )//edit
      {
        int pos = 2;
        float ID = GetFloat(pos);
        bool exist = false;
        int field = 0;
        while(settings::config[field].ID != 0)
        {
          if( settings::config[field].ID == (uInt)ID )
            exist = true;
          field++;
        }
        if( ( Procesor::buff[pos++] == '=' ) && exist )
        {
          float value = GetFloat(pos);
          WriteSection(ID*sizeof(float),value);
        }        
      }
      if( Procesor::buff[1] == 0 )
        Serial.println("[HLP:$$ $# $G $I $N $x=val $Nx=line $J=line $SLP $C $X $H ~ ! ? ctrl-x]");
      if( Procesor::buff[1] == 'I' )
      {
        Serial.println("[VER:1.1g.20180813.Mega:]");
        Serial.println("[OPT:VNM+,35,255]");
      }
      if( Procesor::buff[1] == '#' )
      {
        Serial.println("[G54:0.000,0.000,0.000]");
        Serial.println("[G55:0.000,0.000,0.000]");
        Serial.println("[G57:0.000,0.000,0.000]");
        Serial.println("[G58:0.000,0.000,0.000]");
        Serial.println("[G59:0.000,0.000,0.000]");
        Serial.println("[G28:0.000,0.000,0.000]");
        Serial.println("[G30:0.000,0.000,0.000]");
        Serial.println("[G92:0.000,0.000,0.000]");
        Serial.println("[TLO:0.000]");
        Serial.println("[PRB:0.000,0.000,0.000:0]");
      }
      if( Procesor::buff[1] == 'G' )
      {
        Serial.print("[GC:G0 G54 G17");
        Serial.print((Machine::G90_G91 == 0)?"G90":"G91");
        Serial.print((Machine::G20_G21 == 0)?"G20":"G21");
        Serial.print("G94");
        if( Machine::M3_M5 == 0 )
          Serial.print("M5");
        if( Machine::M3_M5 == 1 )
          Serial.print("M3");
        if( Machine::M3_M5 == 2 )
          Serial.print("M4");
        Serial.print("M9 T0 S");
        Serial.print(Machine::speed_S);
        Serial.print(" F");
        Serial.print(Machine::speed_F);
        Serial.println("]");
      }
      //cofig stuff here
      //$J= jog funcions
      if( ( Procesor::buff[1] == 'J' ) && ( Procesor::buff[2] == '=' ) )
      {
        //$J=G21G91X0.000Y0.000Z-1.000F1000
        Procesor::sFrame * frame = ProcessFrame(3);
        //Fid G commands list
        Procesor::Gcode_list::count=0;
        
        for(int i = 0; i < frame->count; i++ )
        {
          if( frame->Unit[i].Letter == 'G' )
          {
            Procesor::Gcode_list::list[Procesor::Gcode_list::count] = frame->Unit[i].value;
            Procesor::Gcode_list::count++;
          }
        }
          //float Gcode_num = Procesor::Gcode_list::list[i];
        float * f;
        f = FindInFrame('X');
          Procesor::Xvalue = (f != 0 )?f[0]:Machine::pos_X;
        f = FindInFrame('Y');
          Procesor::Yvalue = (f != 0 )?f[0]:Machine::pos_Y;
        f = FindInFrame('Z');
          Procesor::Zvalue = (f != 0 )?f[0]:Machine::pos_Z;
        f = FindInFrame('A');
          Procesor::Avalue = (f != 0 )?f[0]:Machine::pos_A;
        f = FindInFrame('B');
          Procesor::Bvalue = (f != 0 )?f[0]:Machine::pos_B;
        f = FindInFrame('U');
          Procesor::Avalue = (f != 0 )?f[0]:Machine::pos_A;
        f = FindInFrame('V');
          Procesor::Bvalue = (f != 0 )?f[0]:Machine::pos_B;
        f = FindInFrame('F');
          Procesor::Fvalue = (f != 0 )?f[0]:Machine::speed_F;
        float temp_G20_G21 = Machine::G20_G21;
        float temp_G90_G91 = Machine::G90_G91;
        for( int i = 0; i < Procesor::Gcode_list::count; i++ )//inch
        {
          float G = Procesor::Gcode_list::list[i];
          if( G == 20 )
            Machine::G20_G21 = 0;
          if( G == 21 )
            Machine::G20_G21 = 1;
        }
        for( int i = 0; i < Procesor::Gcode_list::count; i++ )
        {
          float G = Procesor::Gcode_list::list[i];
          if( G == 90 )
            Machine::G90_G91 = 0;
          if( G == 91 )
            Machine::G90_G91 = 1;
        }
        DoMove();
        Machine::G20_G21 = temp_G20_G21;
        Machine::G90_G91 = temp_G90_G91;
        sei();
        Machine::machine_status = 1;
      }
    }
    Serial.println("ok");
  }
  if( Procesor::data_ready && ( Machine::machine_status == 0 ) )
  {    
    Procesor::data_ready = false;
    Procesor::len = Procesor::buff;
    Procesor::buff_pos=0;
    Procesor::sFrame * frame = ProcessFrame();

    //Fid G commands list
    Procesor::Gcode_list::count=0;
    for(int i = 0; i < frame->count; i++ )
    {
      if( frame->Unit[i].Letter == 'G' )
      {
        Procesor::Gcode_list::list[Procesor::Gcode_list::count] = frame->Unit[i].value;
        Procesor::Gcode_list::count++;
      }
    }
    for( int i = 0; i < Procesor::Gcode_list::count; i++)
    {
      float Gcode_num = Procesor::Gcode_list::list[i];
      float * f;
      f = FindInFrame('X');
      Procesor::Xvalue = (f != 0 )?f[0]:Machine::pos_X;
      f = FindInFrame('Y');
      Procesor::Yvalue = (f != 0 )?f[0]:Machine::pos_Y;
      f = FindInFrame('Z');
      Procesor::Zvalue = (f != 0 )?f[0]:Machine::pos_Z;

      f = FindInFrame('A');
      Procesor::Avalue = (f != 0 )?f[0]:Machine::pos_A;
      f = FindInFrame('B');
      Procesor::Bvalue = (f != 0 )?f[0]:Machine::pos_B;

      f = FindInFrame('U');
      Procesor::Avalue = (f != 0 )?f[0]:Machine::pos_A;
      f = FindInFrame('V');
      Procesor::Bvalue = (f != 0 )?f[0]:Machine::pos_B;

      f = FindInFrame('K');
      Procesor::Kvalue = (f != 0 )?f[0]:0;
      f = FindInFrame('F');
      Procesor::Fvalue = (f != 0 )?f[0]:Machine::speed_F;
      //do Gcode
      if( Gcode_num == 92 )//G92 zero axis XZ
      {
        Machine::pos_X = Procesor::Xvalue;
        Machine::pos_Y = Procesor::Yvalue;
        Machine::pos_Z = Procesor::Zvalue;
        Machine::pos_A = Procesor::Avalue;
        Machine::pos_B = Procesor::Bvalue;
      }
      if( Gcode_num == 20 )
        Machine::G20_G21 = 0;
      if( Gcode_num == 21 )
        Machine::G20_G21 = 1;
      if( Gcode_num == 90 )
        Machine::G90_G91 = 0;
      if( Gcode_num == 91 )
        Machine::G90_G91 = 1;
      if( Gcode_num == 4 )//wait
      {
        Machine::time = Procesor::Pvalue + Procesor::Svalue*1000;
        Machine::timems = millis();
        Machine::delay_time = true;
        Machine::machine_status = 1;
      }
      if( Gcode_num == 33 )//G92 zero axis XZ
      {
        if( Procesor::Kvalue < 0.01F )
        {
          Serial.println("Error: K incorect or negative\n");
          return;
        }
        int move_Z;
        if( Machine::G20_G21 == 0 )
          Procesor::Zvalue *= 25.4;
        if( Machine::G90_G91 == 0 )
          move_Z = (int)((Machine::pos_Z - Procesor::Zvalue)* (float)driver_Z::STEP_PERMM);
        if( Machine::G90_G91 == 1 )
          move_Z = (int)((Procesor::Zvalue)* (float)driver_Z::STEP_PERMM);
        digitalWrite(driver_Z::DIRECTION, ( move_Z < 0 ) ^ ( driver_Z::FLIP == true ));
        
        //get spindle speed rpm
        int speed_in_rps = 0;
        bool last = false;
        unsigned long tim = millis();
        while( tim + 1000 >= millis())
        {
          if( !last )
          {
            if( digitalRead(spindle::SYNC) == HIGH )
            {
              speed_in_rps++;
              last =!last;
            }
          }
          else
          {
            if( digitalRead(spindle::SYNC) == LOW )
              last =!last;
          }
        }
        speed_in_rps/=spindle::SYNC_PULSES;
        if( speed_in_rps < 1 )
        {
          Serial.println("Error: speed to low or no sync pulses\n");
          return;
        }
        Procesor::Fvalue = Procesor::Kvalue*speed_in_rps*60;
        
        int pulses_per_sec_Z = (int)((Procesor::Fvalue/60.0F)*(float)driver_Z::STEP_PERMM);
        TIMERS::counterX_target = abs(move_Z);
        TCCR2B = (1<<WGM12) | (1<<CS22)| (1<<CS21); 
        TIMSK2 = (1<<OCIE1A);
        OCR2A = 31250/pulses_per_sec_Z;
        //sync with spindle
        while( digitalRead(spindle::INDEX) == LOW )
          delay(1);
        while( digitalRead(spindle::INDEX) == HIGH )
          delay(1);
        while( digitalRead(spindle::INDEX) == LOW )
          delay(1);
        
        Machine::machine_status = 2;
        Machine::sub_mode = 33;//G33
      }
      if(( Gcode_num == 0 ) || ( Gcode_num == 1 ))//G0/G1 move XZ
      {
        DoMove();
        Machine::machine_status = 2;
        Machine::sub_mode = 1;//G33
      }
    }
    Serial.println("ok");
  }
}

float GetFloat(int &pos)
{
  unsigned int str_length = Procesor::len;
  float value = 0;
  bool point = false;
  float poin = 1;
  float sign = 1;
  for( ;pos < str_length; pos++)
  {
    unsigned int ch = Procesor::buff[pos];
    if(ch == '-' )
      sign = -1;
    if(( ch > 0x60 ) && ( ch < 0x7B ))
      return value*sign;
    if(( ch > 0x40 ) && ( ch < 0x5B ))
      return value*sign;
    if(ch == ' ' )
      return value*sign;
    if(ch == '.' )
        point = true;
    if((ch >= '0' ) && (ch <= '9' ))
      if( point )
      {
        poin*=10;
        value += (ch-'0')/poin;
      }
      else
      {
        value *= 10;
        value += (ch-'0');
      }
  }
  return value*sign;
}

//driver_X
ISR(TIMER0_COMPA_vect){
  if( TIMERS::onoffX )
    digitalWrite(driver_X::STEP,HIGH);
  else
    digitalWrite(driver_X::STEP,LOW);
  if( TIMERS::onoffX )
  {
    TIMERS::counterX_target--;
    TIMERS::counterX_done++;
  }
  TIMERS::onoffX = !TIMERS::onoffX;
  if( TIMERS::counterX_target <= 0 )
  {
    TIMERS::counterX_target = 0;
    TCCR0B = 0;
  }
  TCNT0  = 0;
}

//driver_Y
ISR(TIMER1_COMPA_vect){
  if( TIMERS::onoffY )
    digitalWrite(driver_Y::STEP,HIGH);
  else
    digitalWrite(driver_Y::STEP,LOW);
  if( TIMERS::onoffY )
  {
    TIMERS::counterY_target--;
    TIMERS::counterY_done++;
  }
  TIMERS::onoffY = !TIMERS::onoffY;
  if( TIMERS::counterY_target <= 0 )
  {
    TIMERS::counterY_target = 0;
    TCCR1B = 0;
  }
  TCNT1  = 0;
}

//driver_Z
ISR(TIMER2_COMPA_vect){
  
  if( TIMERS::onoffZ )
    digitalWrite(driver_Z::STEP,HIGH);
  else
    digitalWrite(driver_Z::STEP,LOW);
  if( TIMERS::onoffZ )
  {
    TIMERS::counterZ_target--;
    TIMERS::counterZ_done++;
  }
  TIMERS::onoffZ = !TIMERS::onoffZ;
  if( TIMERS::counterZ_target <= 0 )
  {
    TIMERS::counterZ_target = 0;
    TCCR2B = 0;
  }
  TCNT2  = 0;
}

//driver_A
ISR(TIMER3_COMPA_vect){
  
  if( TIMERS::onoffA )
    digitalWrite(driver_A::STEP,HIGH);
  else
    digitalWrite(driver_A::STEP,LOW);
  if( TIMERS::onoffA )
  {
    TIMERS::counterA_target--;
    TIMERS::counterA_done++;
  }
  TIMERS::onoffA = !TIMERS::onoffA;
  if( TIMERS::counterA_target <= 0 )
  {
    TIMERS::counterA_target = 0;
    TCCR3B = 0;
  }
  TCNT3  = 0;
}

//driver_B
ISR(TIMER4_COMPA_vect){
  
  if( TIMERS::onoffB )
    digitalWrite(driver_B::STEP,HIGH);
  else
    digitalWrite(driver_B::STEP,LOW);
  if( TIMERS::onoffB )
  {
    TIMERS::counterB_target--;
    TIMERS::counterB_done++;
  }
  TIMERS::onoffB = !TIMERS::onoffB;
  if( TIMERS::counterB_target <= 0 )
  {
    TIMERS::counterB_target = 0;
    TCCR4B = 0;
  }
  TCNT4  = 0;
}

//spindle
ISR(TIMER5_COMPA_vect){
  
  if( TIMERS::spindle_up_down )
    digitalWrite(spindle::PWM,HIGH);
  else
    digitalWrite(spindle::PWM,LOW);
  if( TIMERS::spindle_up_down )
    OCR5A = 100-TIMERS::spindle_speed;
  else
    OCR5A = TIMERS::spindle_speed;
  TIMERS::spindle_up_down = !TIMERS::spindle_up_down;
}

float* FindInFrame(char letter)
{
  for(int i = 0; i < Procesor::Frame.count; i++ )
  {
    if( Procesor::Frame.Unit[i].Letter == letter )
    {
      return &Procesor::Frame.Unit[i].value;
    }
  }
  return 0;
}

Procesor::sFrame* ProcessFrame(int pos)
{
  Procesor::Frame.count = 0;
  bool end = false;
  int Letters_len = strlen(Procesor::leters_big);
  
  while( ( pos < Procesor::len ) && ( !end ) )
  {
    int pre_pos = pos;
    for( int i = 0; i < Letters_len; i++ )
      if( ( Procesor::buff[pos] == Procesor::leters_big[i] ) || ( Procesor::buff[pos] == Procesor::leters_small[i] ) )
      {
        Procesor::Frame.Unit[Procesor::Frame.count].Letter = Procesor::leters_big[i]; 
        Procesor::Frame.Unit[Procesor::Frame.count].value  = GetFloat(++pos);
        Procesor::Frame.count++;
      }
    while( ( Procesor::buff[pos] == ' ' ) || ( Procesor::buff[pos] == 9 ) )//space/tab
        pos++;
    if( pre_pos == pos )//no more procesing
      end = true;
  }
  return &Procesor::Frame;
}

void DoMove()
{
  int move_X,move_Y,move_Z;
  int move_A,move_B,move_C;
  if( Machine::G20_G21 == 0 )
  {
    Procesor::Xvalue *= 25.4;
    Procesor::Yvalue *= 25.4;
    Procesor::Zvalue *= 25.4;
    Procesor::Avalue *= 25.4;
    Procesor::Bvalue *= 25.4;
    Procesor::Cvalue *= 25.4;
  }
  if( Machine::G90_G91 == 0 )
  {
    move_X = (int)((Machine::pos_X - Procesor::Xvalue)* (float)driver_X::STEP_PERMM);
    move_Y = (int)((Machine::pos_Y - Procesor::Yvalue)* (float)driver_Y::STEP_PERMM);
    move_Z = (int)((Machine::pos_Z - Procesor::Zvalue)* (float)driver_Z::STEP_PERMM);
    move_A = (int)((Machine::pos_A - Procesor::Avalue)* (float)driver_A::STEP_PERANG);
    move_B = (int)((Machine::pos_B - Procesor::Bvalue)* (float)driver_B::STEP_PERANG);
  }
  if( Machine::G90_G91 == 1 )
  {
    move_X = (int)((Procesor::Xvalue)* (float)driver_X::STEP_PERMM);
    move_Y = (int)((Procesor::Yvalue)* (float)driver_Y::STEP_PERMM);
    move_Z = (int)((Procesor::Zvalue)* (float)driver_Z::STEP_PERMM);
    move_A = (int)((Procesor::Avalue)* (float)driver_A::STEP_PERANG);
    move_B = (int)((Procesor::Bvalue)* (float)driver_B::STEP_PERANG);
  }
  digitalWrite(driver_X::DIRECTION, ( move_X < 0 ) ^ ( driver_X::FLIP == true ));
  digitalWrite(driver_Y::DIRECTION, ( move_Y < 0 ) ^ ( driver_Y::FLIP == true ));
  digitalWrite(driver_Z::DIRECTION, ( move_Z < 0 ) ^ ( driver_Z::FLIP == true ));
  digitalWrite(driver_A::DIRECTION, ( move_A < 0 ) ^ ( driver_A::FLIP == true ));
  digitalWrite(driver_B::DIRECTION, ( move_B < 0 ) ^ ( driver_B::FLIP == true ));

  int pulses_per_sec_X = (int)((Procesor::Fvalue/60.0F)*(float)driver_X::STEP_PERMM);
  int pulses_per_sec_Y = (int)((Procesor::Fvalue/60.0F)*(float)driver_Y::STEP_PERMM);
  int pulses_per_sec_Z = (int)((Procesor::Fvalue/60.0F)*(float)driver_Z::STEP_PERMM);
  int pulses_per_sec_A = (int)((Procesor::Fvalue/60.0F)*(float)driver_A::STEP_PERANG);
  int pulses_per_sec_B = (int)((Procesor::Fvalue/60.0F)*(float)driver_B::STEP_PERANG);
  //sychronization:
  float X_deest = fabsf((float)move_X)/(float)pulses_per_sec_X;
  float Y_deest = fabsf((float)move_Y)/(float)pulses_per_sec_Y;
  float Z_deest = fabsf((float)move_Z)/(float)pulses_per_sec_Z;
  float A_deest = fabsf((float)move_A)/(float)pulses_per_sec_A;
  float B_deest = fabsf((float)move_B)/(float)pulses_per_sec_B;
  
  if(( fabs(move_X) >0.01F ) + ( fabs(move_Y) >0.01F  ) + ( fabs(move_Z) >0.01F  ) + ( fabs(move_A) >0.01F ) + ( fabs(move_B) >0.01F  ) > 1 )
  {
    if(( X_deest > Y_deest ) && ( X_deest > Z_deest )&&( X_deest > A_deest ) && ( X_deest > B_deest ))
    {
      if( fabs(move_Z) >0.01F  )
        pulses_per_sec_Z = (((float)pulses_per_sec_Z) / (X_deest/Z_deest));
      if( fabs(move_Y) >0.01F  )
        pulses_per_sec_Y = (((float)pulses_per_sec_Y) / (X_deest/Y_deest));
      if( fabs(move_A) >0.01F  )
        pulses_per_sec_A = (((float)pulses_per_sec_A) / (X_deest/A_deest));
      if( fabs(move_B) >0.01F  )
        pulses_per_sec_B = (((float)pulses_per_sec_B) / (X_deest/B_deest));
    }
    if(( Y_deest > X_deest ) && ( Y_deest > Y_deest )&&( Y_deest > A_deest ) && ( Y_deest > B_deest ))
    {
      if( fabs(move_X) >0.01F  )
        pulses_per_sec_X = (((float)pulses_per_sec_X) / (Y_deest/X_deest));
      if( fabs(move_Z) >0.01F  )
        pulses_per_sec_Z = (((float)pulses_per_sec_Z) / (Y_deest/Z_deest));
      if( fabs(move_A) >0.01F  )
        pulses_per_sec_A = (((float)pulses_per_sec_A) / (Y_deest/A_deest));
      if( fabs(move_B) >0.01F  )
        pulses_per_sec_B = (((float)pulses_per_sec_B) / (Y_deest/B_deest));
    }
    if(( Z_deest > X_deest ) && ( Z_deest > Y_deest )&&( Z_deest > A_deest ) && ( Z_deest > B_deest ))
    {
      if( fabs(move_X) >0.01F  )
        pulses_per_sec_X = (((float)pulses_per_sec_X) / (Z_deest/X_deest));
      if( fabs(move_Y) >0.01F  )
        pulses_per_sec_Y = (((float)pulses_per_sec_Y) / (Z_deest/Y_deest));
      if( fabs(move_A) >0.01F  )
        pulses_per_sec_A = (((float)pulses_per_sec_A) / (Z_deest/A_deest));
      if( fabs(move_B) >0.01F  )
        pulses_per_sec_B = (((float)pulses_per_sec_B) / (Z_deest/B_deest));
    }
    if(( A_deest > X_deest ) && ( A_deest > Y_deest )&&( A_deest > Z_deest ) && ( A_deest > B_deest ))
    {
      if( fabs(move_X) >0.01F  )
        pulses_per_sec_X = (((float)pulses_per_sec_X) / (A_deest/X_deest));
      if( fabs(move_Y) >0.01F  )
        pulses_per_sec_Y = (((float)pulses_per_sec_Y) / (A_deest/Y_deest));
      if( fabs(move_Z) >0.01F  )
        pulses_per_sec_Z = (((float)pulses_per_sec_Z) / (A_deest/Z_deest));
      if( fabs(move_B) >0.01F  )
        pulses_per_sec_B = (((float)pulses_per_sec_B) / (A_deest/B_deest));
    }
    if(( B_deest > X_deest ) && ( B_deest > Y_deest )&&( B_deest > Z_deest ) && ( B_deest > A_deest ))
    {
      if( fabs(move_X) >0.01F  )
        pulses_per_sec_X = (((float)pulses_per_sec_X) / (B_deest/X_deest));
      if( fabs(move_Y) >0.01F  )
        pulses_per_sec_Y = (((float)pulses_per_sec_Y) / (B_deest/Y_deest));
      if( fabs(move_Z) >0.01F  )
        pulses_per_sec_Z = (((float)pulses_per_sec_Z) / (B_deest/Z_deest));
      if( fabs(move_A) >0.01F  )
        pulses_per_sec_A = (((float)pulses_per_sec_A) / (B_deest/A_deest));
    }
  }
  TIMERS::counterX_target = abs(move_X);
  TIMERS::counterY_target = abs(move_Y);
  TIMERS::counterZ_target = abs(move_Z);
  TIMERS::counterA_target = abs(move_A);
  TIMERS::counterB_target = abs(move_B);
  TIMERS::counterX_done = 0;
  TIMERS::counterY_done = 0;
  TIMERS::counterZ_done = 0;
  TIMERS::counterA_done = 0;
  TIMERS::counterB_done = 0;
  cli();
  TCCR0B = (1<<WGM12) | (1<<CS12); 
  TIMSK0 = (1<<OCIE1A);
  OCR0A = 31250/pulses_per_sec_X;

  TCCR1B = (1<<WGM12) | (1<<CS12);
  TIMSK1 = (1<<OCIE1A);
  OCR1A = 31250/pulses_per_sec_Y;

  TCCR2B = (1<<WGM12) | (1<<CS22)| (1<<CS21);
  TIMSK2 = (1<<OCIE1A);
  OCR2A = 31250/pulses_per_sec_Z;

  TCCR3B = (1<<WGM12) | (1<<CS12); 
  TIMSK3 = (1<<OCIE1A);
  OCR3A = 31250/pulses_per_sec_A;

  TCCR4B = (1<<WGM12) | (1<<CS12); 
  TIMSK4 = (1<<OCIE1A);
  OCR4A = 31250/pulses_per_sec_B;
  sei();  
}

void WriteSection(uInt address,float data)
{
  for( int i = 0; i < sizeof(float); i++ )
    EEPROM.write(address+i,((uint8_t*)(void*)&data)[i]);
}

float ReadSection(uInt address)
{
  float f=0;
  for( int i = 0; i < sizeof(float); i++ )
    ((uint8_t*)(void*)&f)[i] = EEPROM.read(address+i);
  return f;
}
